package com.urbancode.air.plugin.command;

import java.io.File;
import java.io.FileInputStream;
import java.net.URI;
import java.nio.charset.Charset;
import java.util.List;
import java.util.Properties;

import org.apache.log4j.PropertyConfigurator;

import com.urbancode.commons.fileutils.FileUtils;
import com.urbancode.ud.client.ComponentClient;
import com.urbancode.ud.client.VersionClient;

public class ImportVersion {

    //**********************************************************************************************
    // CLASS
    //**********************************************************************************************

    //**********************************************************************************************
    // INSTANCE
    //**********************************************************************************************
    private Properties inputProperties;
    private Properties outputProperties;
    
    /**
     * Standard plugin command constructor. It must take two Properties objects. inputProperties will
     * contain all of the values passed into this plugin step. outputProperties can be populated with
     * values and will be written to the output properties file on returning from execute.
     * 
     * @param inputProperties
     * @param outputProperties
     */
    public ImportVersion(Properties inputProperties, Properties outputProperties) {
        this.inputProperties = inputProperties;
        this.outputProperties = outputProperties;
    }
    
    /**
     * Commands must have an execute() method which returns an int. Returning 0 should mean success,
     * anything else means failure.
     */
    public int execute() {
        int returnCode = 0;
        
        File repoDir = null;
        try {
            
            //So log4j doesn't throw warnings during execution.
            Properties props = new Properties();
            props.load(new FileInputStream(new File(System.getenv("AGENT_HOME"), "conf/agent/log4j.properties")));
            PropertyConfigurator.configure(props);

            Properties agentProps = new Properties();
            agentProps.load(new FileInputStream(new File(System.getenv("AGENT_HOME"), "conf/agent/installed.properties")));
            String charsetName = agentProps.getProperty("system.default.encoding");
            Charset charset = null;
            if (charsetName != null) {
                charset = Charset.forName(charsetName);
            }

            String extensionString = inputProperties.getProperty("extensions");
            String[] extensions = new String[0];
            if (extensionString != null ) {
                extensions = extensionString.split(",");
                for (int i=0; i<extensions.length; i++) {
                    extensions[i] = extensions[i].trim();
                }
            }

            String componentName = inputProperties.getProperty("componentName");
            String port = inputProperties.getProperty("port");
            String template = inputProperties.getProperty("template");
            String perforceUsername = inputProperties.getProperty("user");
            String perforcePassword = inputProperties.getProperty("password");
            String p4Path = inputProperties.getProperty("p4path");
            boolean doLoginAndLogout = Boolean.valueOf(inputProperties.getProperty("doLogin"));
            boolean saveExecutePermissions = Boolean.valueOf(inputProperties.getProperty("saveFileExecuteBits"));
      
            String[] includes = {"**/*"};
            String[] excludes = {};
            boolean isUseVFS = Boolean.valueOf(inputProperties.getProperty("isUseVFS"));
            
            String username = "PasswordIsAuthToken";
            String password = String.format("{\"token\": \"%s\"}", System.getenv("AUTH_TOKEN"));
            String webUrl = System.getenv("AH_WEB_URL");
            URI url = new URI(webUrl);
            
            VersionClient versionClient = new VersionClient(url, username, password);
            ComponentClient componentClient = new ComponentClient(url, username, password);
            PerforceClient perforceClient = new PerforceClient(p4Path, port, perforceUsername, perforcePassword, template, doLoginAndLogout);
            
            String p4Version = perforceClient.getLatestVersion();
            List<String> versions = componentClient.getComponentVersions(componentName, false);
            boolean hasVersion = versions.contains(p4Version);
            System.out.println(String.format("Version from perforce: %s", p4Version));
            
            if (!hasVersion) {
                System.out.println(String.format("Creating version: %s", p4Version));
                String versionId = versionClient.createVersion(componentName, p4Version, " ").toString();
                outputProperties.setProperty("VersionID", versionId);
                
                if (isUseVFS) {
                    repoDir = perforceClient.downloadFiles(p4Version);
                    System.out.println("Uploading artifact files.");
                    versionClient.addVersionFiles(componentName, p4Version, repoDir, "",
                            includes, excludes, saveExecutePermissions, true, charset, extensions);
                }
                else {
                    System.out.println("Skipping Artifact Upload: Component is configured not to use codestation.");
                }
            }
            else {
                System.out.println(String.format("Component Already Has Version: %s", p4Version));
            }
        }
        catch (Exception e) {
            System.err.println(String.format("Error Creating New Version: %s", e.getMessage()));
            returnCode = 1;
        }
        finally {
            if (repoDir != null && repoDir.exists()) {
                FileUtils.deleteFile(repoDir);
            }
        }
        
        return returnCode;
    }
}
