/*
* Licensed Materials - Property of IBM Corp.
* IBM UrbanCode Deploy
* (c) Copyright IBM Corporation 2011, 2013. All Rights Reserved.
*
* U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by
* GSA ADP Schedule Contract with IBM Corp.
*/
package com.urbancode.air.plugin.command;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

import com.urbancode.commons.util.Check;
import com.urbancode.commons.util.IO;
import com.urbancode.commons.util.processes.Processes;

public class PerforceClient {

    //**********************************************************************************************
    // CLASS
    //**********************************************************************************************
    final static private Logger log = Logger.getLogger(PerforceClient.class);

    //**********************************************************************************************
    // INSTANCE
    //**********************************************************************************************
    final private Processes processes = new Processes();

    final private String p4Path;
    final private String port;
    final private String user;
    final private String password;
    final private String template;
    final private boolean doLogin;
    private File workDir;
    final private String clientName;
    private File p4TicketsFile;
    private List<String> indcludeFiles;
    private List<String> excludeFiles;
    boolean isSetup = false;
    private Pattern excludedFilesPattern;

    //----------------------------------------------------------------------------------------------
    public PerforceClient(String p4Path, String port, String user, String password, String template,
            boolean doLogin) {
        Check.nonBlank(p4Path, "p4Path");
        Check.nonBlank(port, "port");
        Check.nonBlank(user, "user");
        Check.nonBlank(template, "template");
        this.p4Path = p4Path.trim();
        this.port = port.trim();
        this.user = user.trim();
        this.password = trimToNull(password);
        this.template = template.trim();
        this.doLogin = doLogin;
        this.clientName = newClientName();
    }

    //----------------------------------------------------------------------------------------------
    public String getLatestVersion()
    throws IOException, InterruptedException {
        assertSetup();
        Integer latestVersion = Integer.MIN_VALUE;
        if (excludeFiles.isEmpty()){
            for (String filePath : indcludeFiles) {
                Integer version = getLastestVersionForFile(filePath);
                if (version != null && version > latestVersion) {
                    latestVersion = version;
                }
            }
        }
        else {
            List<Integer> versions = getAllVersions();
            latestVersion = getLatestVersionWithChangeNotExcluded(versions);
        }
        return String.valueOf(latestVersion);
    }

    //----------------------------------------------------------------------------------------------
    public File downloadFiles(String version)
    throws IOException, InterruptedException {
        assertSetup();
        List<String> commandLine = new ArrayList<String>();
        commandLine.add(p4Path);
        commandLine.add("-c");
        commandLine.add(clientName);
        commandLine.add("-p");
        commandLine.add(port);
        commandLine.add("-u");
        commandLine.add(user);
        if (password != null && !doLogin) {
            commandLine.add("-P");
            commandLine.add(password);
        }
        commandLine.add("sync");
        commandLine.add("@" + version);
        handleProcess(execute(commandLine), null);
        return workDir;
    }

    //----------------------------------------------------------------------------------------------
    private void setup()
    throws IOException, InterruptedException {
        this.workDir = newPath();
        this.p4TicketsFile = newPath();
        IO.mkdirs(workDir);
        if (doLogin){
            p4TicketsFile.createNewFile();
            login();
        }
        String clientSpec = createClientSpec();
        setDepotFileLists(clientSpec);
        excludedFilesPattern = convertToPattern(excludeFiles);
        clientSpec = changeClientSpecRoot(clientSpec);
        createClient(clientSpec);
    }

    //----------------------------------------------------------------------------------------------
    public void close()
    throws IOException, InterruptedException {
        try {
            removeClient();
            if (doLogin) {
                logout();
                p4TicketsFile.delete();
            }
        }
        finally {
            isSetup = false;
            IO.delete(workDir);
        }
    }

    //----------------------------------------------------------------------------------------------
    private void removeClient()
    throws IOException, InterruptedException {
        List<String> commandLine = new ArrayList<String>();
        commandLine.add(p4Path);
        commandLine.add("-p");
        commandLine.add(port);
        commandLine.add("-u");
        commandLine.add(user);
        if (password != null && !doLogin) {
            commandLine.add("-P");
            commandLine.add(password);
        }
        commandLine.add("client");
        commandLine.add("-d");
        commandLine.add(clientName);
        handleProcess(execute(commandLine), null);
    }

    //----------------------------------------------------------------------------------------------
    private String createClientSpec()
    throws IOException, InterruptedException {
        List<String> commandLine = new ArrayList<String>();
        commandLine.add(p4Path);
        commandLine.add("-p");
        commandLine.add(port);
        commandLine.add("-u");
        commandLine.add(user);
        if (password != null && !doLogin) {
            commandLine.add("-P");
            commandLine.add(password);
        }
        commandLine.add("client");
        commandLine.add("-o");
        commandLine.add("-t");
        commandLine.add(template);
        commandLine.add(clientName);
        return handleProcess(execute(commandLine), null);
    }

    //----------------------------------------------------------------------------------------------
    private void createClient(String clientSpec)
    throws IOException, InterruptedException {
        List<String> commandLine = new ArrayList<String>();
        commandLine.add(p4Path);
        commandLine.add("-p");
        commandLine.add(port);
        commandLine.add("-u");
        commandLine.add(user);
        if (password != null && !doLogin) {
            commandLine.add("-P");
            commandLine.add(password);
        }
        commandLine.add("client");
        commandLine.add("-i");
        handleProcess(execute(commandLine), clientSpec);
    }

    //----------------------------------------------------------------------------------------------
    private String changeClientSpecRoot(String clientSpec) {
        StringBuilder builder = new StringBuilder();
        String[] tokens = clientSpec.split("\\r?\\n");
        for (String token : tokens) {
            if (token.startsWith("Root:")) {
                builder.append("Root:  ");
                builder.append(workDir.getAbsolutePath());
                builder.append("\n");
            }
            else {
                builder.append(token);
                builder.append("\n");
            }
        }
        return builder.toString();
    }

    //----------------------------------------------------------------------------------------------
    public void setDepotFileLists(String clientSpec) {
        Pattern includesFilePattern = Pattern.compile("^\\+?(\"?)\\+?(\\/\\/.*?) \"?\\/\\/.*$");
        Pattern excludesFilePattern = Pattern.compile("^(\"?)-(\\/\\/.*?) \"?\\/\\/.*$");
        StringBuilder builder;
        List<String> excludeFiles = new ArrayList<String>();
        List<String> includeFiles = new ArrayList<String>();
        List<String> mappingLines = getViewMappingFromClient(clientSpec);
        for (String mapping : mappingLines){
            builder = new StringBuilder();
            Matcher matcher = includesFilePattern.matcher(mapping);
            if(matcher.find()){
                for (int i =1 ; i <= matcher.groupCount(); i++) {
                    builder.append(matcher.group(i));
                }
                includeFiles.add(builder.toString());
            }
            matcher = excludesFilePattern.matcher(mapping);
            if (matcher.find()){
                for (int i =1 ; i <= matcher.groupCount(); i++) {
                    builder.append(matcher.group(i));
                }
                excludeFiles.add(builder.toString());
            }
        }
        this.excludeFiles = excludeFiles;
        this.indcludeFiles = includeFiles;
    }

    //----------------------------------------------------------------------------------------------
    public List<String> getViewMappingFromClient(String clientSpec) {
        boolean insideViewSection = false;
        String[] lines = clientSpec.split("\n");
        List<String> mappingLines = new ArrayList<String>();
        for (int i=0; i<lines.length; i++) {
            String line = lines[i].trim();
            if (line.startsWith("#")){
                continue;
            }
            else if (insideViewSection){
                mappingLines.add(line);
            }
            else if (line.startsWith("View:")){
                insideViewSection = true;
            }
        }
        return mappingLines;
    }

    //----------------------------------------------------------------------------------------------
    private void login()
    throws IOException, InterruptedException {
        List<String> commandLine = new ArrayList<String>();
        commandLine.add(p4Path);
        commandLine.add("-p");
        commandLine.add(port);
        commandLine.add("-u");
        commandLine.add(user);
        commandLine.add("login");
        handleProcess(execute(commandLine), password);
    }

    //----------------------------------------------------------------------------------------------
    private void logout()
    throws IOException, InterruptedException {
        List<String> commandLine = new ArrayList<String>();
        commandLine.add(p4Path);
        commandLine.add("-p");
        commandLine.add(port);
        commandLine.add("-u");
        commandLine.add(user);
        commandLine.add("logout");
        handleProcess(execute(commandLine), null);
    }

    //----------------------------------------------------------------------------------------------
    private String handleProcess(Process process, String input)
    throws IOException, InterruptedException {
        String result;
        try {
            Future<String> errorFuture = processes.getErrorStringFuture(process);
            Future<String> outputFuture = processes.getOutputStringFuture(process);
            if (input != null && input.length() > 0) {
                process.getOutputStream().write(input.getBytes());
            }
            process.getOutputStream().close();
            int exitCode = process.waitFor();
            if (exitCode != 0) {
                String errorText = "";
                try {
                    errorText = errorFuture.get();
                }
                catch (ExecutionException swallow) {
                }
                throw new IOException(String.format(
                    "Command failed with code %d: %s",
                    exitCode,
                    errorText));
            }
            try {
                result = outputFuture.get();
            }
            catch (ExecutionException e) {
                throw (IOException) new IOException().initCause(e.getCause());
            }
            if (log.isDebugEnabled()) {
                try {
                    log.debug("Command output: " + outputFuture.get());
                    log.debug("Command error: " + errorFuture.get());
                }
                catch (ExecutionException swallow) {
                }
            }
        }
        finally {
            process.destroy();
        }
        return result;
    }

    //----------------------------------------------------------------------------------------------
    private Process execute(List<String> commandLine)
    throws IOException {
        if (log.isDebugEnabled()) {
            log.debug("Command Line :" + commandLine.toString().replace(password, "****"));
        }
        ProcessBuilder builder = new ProcessBuilder(commandLine);
        if (p4TicketsFile != null) {
            builder.environment().put("P4TICKETS", p4TicketsFile.getAbsolutePath());
            if (log.isDebugEnabled()) {
                log.debug("Set P4TICKETS=" + p4TicketsFile.getAbsolutePath());
            }
        }
        builder.directory(workDir);
        return builder.start();
    }

    //----------------------------------------------------------------------------------------------
    private File newPath()
    throws IOException {
        return new File(
            System.getProperty("java.io.tmpdir"),
            UUID.randomUUID().toString());
    }

    //----------------------------------------------------------------------------------------------
    private String newClientName() {
        return "IBM_UCD-" + UUID.randomUUID().toString();
    }

    //----------------------------------------------------------------------------------------------
    private String trimToNull(String text) {
        String result = text;
        if (result != null) {
            result = result.trim();
            if (result.length() == 0) {
                result = null;
            }
        }
        return result;
    }

    //----------------------------------------------------------------------------------------------
    private void assertSetup()
    throws IOException, InterruptedException {
        if (!isSetup) {
            setup();
            isSetup = true;
        }
    }

    //----------------------------------------------------------------------------------------------
    private Integer getLastestVersionForFile(String file)
    throws IOException, InterruptedException {
        return getChangesForFile(file, 1).get(0);
    }

    //----------------------------------------------------------------------------------------------
    private List<Integer> getChangesForFile(String file, int numberOfChanges)
    throws IOException, InterruptedException {
        ArrayList<Integer> result = new ArrayList<Integer>();
        List<String> commandLine = new ArrayList<String>();
        commandLine.add(p4Path);
        commandLine.add("-p");
        commandLine.add(port);
        commandLine.add("-u");
        commandLine.add(user);
        if (password != null && !doLogin) {
            commandLine.add("-P");
            commandLine.add(password);
        }
        commandLine.add("changes");
        if (numberOfChanges > 0) {
            commandLine.add("-m");
            commandLine.add("1");
        }
        commandLine.add(file);
        String output = handleProcess(execute(commandLine), null);
        String[] changeList = output.split("\\n");
        for (String change : changeList){
            Pattern pattern = Pattern.compile("^Change (\\d+) ");
            Matcher matcher = pattern.matcher(change);
            if (matcher.find()) {
                result.add(Integer.parseInt(matcher.group(1)));
            }
        }
        return result;
    }

    //----------------------------------------------------------------------------------------------
    public List<String> getIncludePaths(){
        return indcludeFiles;
    }

    //----------------------------------------------------------------------------------------------
    public List<String> getExcludePaths(){
        return excludeFiles;
    }

    //----------------------------------------------------------------------------------------------
    public List<String> clone(List<String> list){
        List<String> clone = new ArrayList<String>();
        for (int i = 0; i< list.size(); i++) {
            clone.add(list.get(i));
        }
        return clone;
    }

    //----------------------------------------------------------------------------------------------
    public List<Integer> getAllVersions()
    throws IOException, InterruptedException {
        List<Integer> versionList = new ArrayList<Integer>();
        assertSetup();
        Set<Integer> versions = new HashSet<Integer>();
        for (String file : indcludeFiles) {
            versions.addAll(getChangesForFile(file, 0));
        }
        versionList.addAll(versions);
        Collections.sort(versionList);
        return versionList;
    }

    //----------------------------------------------------------------------------------------------
    private Integer getLatestVersionWithChangeNotExcluded(List<Integer> versions)
    throws IOException, InterruptedException {
        Integer version = versions.remove(versions.size()-1);
        while (!changesExist(version)) {
            if (versions.isEmpty()){
                throw new IOException("No changes exist in this perforce template");
            }
            version = versions.remove(versions.size()-1);
        }
        return version;
    }

    //----------------------------------------------------------------------------------------------
    public boolean changesExist (Integer version)
    throws IOException, InterruptedException{
        String description = getChangeDescription(version);
        List<String> changedFiles = getChangedFilesFromDescription(description);
        boolean isExcluded = false;
        for (String file : changedFiles) {
            Matcher m = excludedFilesPattern.matcher(file);
            if (!m.matches()){
                isExcluded = true;
                break;
            }
        }
        return isExcluded;
    }

    //----------------------------------------------------------------------------------------------
    public Pattern convertToPattern(List<String> files) {
        StringBuilder builder = new StringBuilder();
        for (int i = 0; i< files.size(); i++) { 
            if (i>0) {
                builder.append("|");
            }
            String pattern = files.get(i);
            pattern.replaceAll("%%[1-9]", "*");
            pattern = pattern.replace("*", "[^\\/^*]*");
            pattern = pattern.replace("...", ".*");
            pattern = pattern.replace("/", "\\/");
            builder.append(pattern);
        }
        return Pattern.compile(builder.toString());
    }

    //----------------------------------------------------------------------------------------------
    public List<String> getChangedFilesFromDescription(String changeDescription){
        boolean inAffectedFilepaths = false;
        Pattern changedFilePattern = Pattern.compile("\\.{3} (.*)#.*");
        Matcher matcher;
        String[] descriptionLines = changeDescription.split("\n");
        List<String> changedFilePaths = new ArrayList<String>();
        for (int i =0; i< descriptionLines.length; i++) {
            String line = descriptionLines[i].trim();
            if (inAffectedFilepaths){
                matcher = changedFilePattern.matcher(line);
                if (matcher.find()) {
                    changedFilePaths.add(matcher.group(1));
                }
                else if (line.isEmpty()){
                    continue;
                }
                else {
                    inAffectedFilepaths = false;
                }
            }
            if (line.contains("Affected files ...")) {
                inAffectedFilepaths = true;
            }
        }
        return changedFilePaths;
    }

    //----------------------------------------------------------------------------------------------
    private String getChangeDescription(Integer change)
    throws IOException, InterruptedException {
        List<String> commandLine = new ArrayList<String>();
        commandLine.add(p4Path);
        commandLine.add("-p");
        commandLine.add(port);
        commandLine.add("-u");
        commandLine.add(user);
        if (password != null && !doLogin) {
            commandLine.add("-P");
            commandLine.add(password);
        }
        commandLine.add("describe");
        commandLine.add("-s");
        commandLine.add(String.valueOf(change));
        return handleProcess(execute(commandLine), null);
    }
}
